from . import SimUtils
from . import SimCanvasRubberband
from SimulatorFramework.TileType import *
from SimulatorFramework.TileAssembly import *
import wx
import wx.richtext as rt
import wx.aui
import wx.lib.agw.cubecolourdialog as CCD
import os
import random
import ast
import sys
if sys.version_info[0] < 3:
    from wx.lib.pubsub import pub
else:
    from pubsub import pub
import wx.stc
import wx.lib.scrolledpanel as scrolled
#from _ast import Num

CELLSIZE  = 60

SIMFW_MODE_PAN_ZOOM    = 0
SIMFW_MODE_SELECT      = 1
SIMFW_MODE_DRAW        = 2

SIMID_ED_SEARCH        = 1


############################################################################################
#
#
class EditorSearchCtrl(wx.SearchCtrl):
    maxSearches = 5
    
    def __init__(self, parent, frame, id=-1, value="",
                 pos=wx.DefaultPosition, size=wx.DefaultSize, style=0,
                 doSearch=None):
        self.frame = frame
        style |= wx.TE_PROCESS_ENTER
        wx.SearchCtrl.__init__(self, parent, id, value, pos, size, style)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnTextEntered)
        #self.Bind(wx.EVT_MENU_RANGE, self.OnMenuItem, id=1, id2=self.maxSearches)
        self.doSearch = doSearch
        self.searches = []
        
    def OnTextEntered(self, event):
        self.frame.DoSearch(event)

############################################################################################
#
#
TILEPANEL_DEFWIDTH = 230
TILEPANEL_DEFHEIGHT = 600

class TileTypePanel(scrolled.ScrolledPanel):
    
    def __init__(self, frame):
        self.tileType = None
        self.origTileColor = None
        self.origTextColor = None

        self.frame = frame
        scrolled.ScrolledPanel.__init__(self, frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(TILEPANEL_DEFWIDTH,TILEPANEL_DEFHEIGHT), style = wx.TAB_TRAVERSAL)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        #self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.AddSpacer(140)
        
        self.colorSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.colorSizer.AddSpacer(25)
        self.tileColorButton  = wx.Button(self, wx.ID_ANY, "Tile Color", size=(60,25))
        self.tileColorButton.Bind(wx.EVT_BUTTON, self.OnTileColor)
        self.colorSizer.Add(self.tileColorButton)
        self.colorSizer.AddSpacer(10)
        self.textColorButton  = wx.Button(self, wx.ID_ANY, "Text Color", size=(60,25))
        self.textColorButton.Bind(wx.EVT_BUTTON, self.OnTextColor)
        self.colorSizer.Add(self.textColorButton)
        self.mainSizer.Add(self.colorSizer)
        self.mainSizer.AddSpacer(10)

        self.nameSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.nameSizer.Add(wx.StaticText(self, label="Name:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.nameSizer.AddSpacer(5)
        self.nameCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(185,25))
        self.nameSizer.Add(self.nameCtrl)
        self.mainSizer.Add(self.nameSizer)
        self.mainSizer.AddSpacer(10)
        
        self.labelSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.labelSizer.Add(wx.StaticText(self, label="Label:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.labelSizer.AddSpacer(9)
        self.labelCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(185,25))
        self.labelSizer.Add(self.labelCtrl)
        self.mainSizer.Add(self.labelSizer)
        self.mainSizer.AddSpacer(10)

        self.mainSizer.Add(wx.StaticText(self, label="North glue"))
        self.northSizer = wx.BoxSizer(wx.HORIZONTAL)#wx.StaticBoxSizer(orient=wx.HORIZONTAL, parent=self, label="North glue")
        self.northSizer.AddSpacer(5)
        self.northSizer.Add(wx.StaticText(self, label="Strength:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.northSizer.AddSpacer(5)
        self.nStrengthCtrl = wx.lib.intctrl.IntCtrl(self, -1, 99, wx.DefaultPosition, wx.Size(20,20))
        #self.xCtrl.SetToolTip("Current minimum x = " + str(self.parentApp.canvas.assembly.minTiledX) + ", maximum x = " + str(self.parentApp.canvas.assembly.maxTiledX))
        self.nStrengthCtrl.SetNoneAllowed(False)
        self.northSizer.Add(self.nStrengthCtrl, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.northSizer.AddSpacer(10)
        self.northSizer.Add(wx.StaticText(self, label="Label:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.northSizer.AddSpacer(5)
        self.nLabelCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(100,25))
        self.northSizer.Add(self.nLabelCtrl)
        self.mainSizer.Add(self.northSizer)
        self.mainSizer.AddSpacer(10)
        
        self.mainSizer.Add(wx.StaticText(self, label="East glue"))
        self.eastSizer = wx.BoxSizer(wx.HORIZONTAL)#wx.StaticBoxSizer(orient=wx.HORIZONTAL, parent=self, label="East glue")
        self.eastSizer.AddSpacer(5)
        self.eastSizer.Add(wx.StaticText(self, label="Strength:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.eastSizer.AddSpacer(5)
        self.eStrengthCtrl = wx.lib.intctrl.IntCtrl(self, -1, 99, wx.DefaultPosition, wx.Size(20,20))
        #self.xCtrl.SetToolTip("Current minimum x = " + str(self.parentApp.canvas.assembly.minTiledX) + ", maximum x = " + str(self.parentApp.canvas.assembly.maxTiledX))
        self.eStrengthCtrl.SetNoneAllowed(False)
        self.eastSizer.Add(self.eStrengthCtrl, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.eastSizer.AddSpacer(10)
        self.eastSizer.Add(wx.StaticText(self, label="Label:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.eastSizer.AddSpacer(5)
        self.eLabelCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(100,25))
        self.eastSizer.Add(self.eLabelCtrl)
        self.mainSizer.Add(self.eastSizer)
        self.mainSizer.AddSpacer(10)
        
        self.mainSizer.Add(wx.StaticText(self, label="South glue"))
        self.southSizer = wx.BoxSizer(wx.HORIZONTAL)#wx.StaticBoxSizer(orient=wx.HORIZONTAL, parent=self, label="South glue")
        self.southSizer.AddSpacer(5)
        self.southSizer.Add(wx.StaticText(self, label="Strength:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.southSizer.AddSpacer(5)
        self.sStrengthCtrl = wx.lib.intctrl.IntCtrl(self, -1, 99, wx.DefaultPosition, wx.Size(20,20))
        #self.xCtrl.SetToolTip("Current minimum x = " + str(self.parentApp.canvas.assembly.minTiledX) + ", maximum x = " + str(self.parentApp.canvas.assembly.maxTiledX))
        self.sStrengthCtrl.SetNoneAllowed(False)
        self.southSizer.Add(self.sStrengthCtrl, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.southSizer.AddSpacer(10)
        self.southSizer.Add(wx.StaticText(self, label="Label:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.southSizer.AddSpacer(5)
        self.sLabelCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(100,25))
        self.southSizer.Add(self.sLabelCtrl)
        self.mainSizer.Add(self.southSizer)
        self.mainSizer.AddSpacer(10)
        
        self.mainSizer.Add(wx.StaticText(self, label="West glue"))
        self.westSizer = wx.BoxSizer(wx.HORIZONTAL)#wx.StaticBoxSizer(orient=wx.HORIZONTAL, parent=self, label="West glue")
        self.westSizer.AddSpacer(5)
        self.westSizer.Add(wx.StaticText(self, label="Strength:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.westSizer.AddSpacer(5)
        self.wStrengthCtrl = wx.lib.intctrl.IntCtrl(self, -1, 99, wx.DefaultPosition, wx.Size(20,20))
        #self.xCtrl.SetToolTip("Current minimum x = " + str(self.parentApp.canvas.assembly.minTiledX) + ", maximum x = " + str(self.parentApp.canvas.assembly.maxTiledX))
        self.wStrengthCtrl.SetNoneAllowed(False)
        self.westSizer.Add(self.wStrengthCtrl, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.westSizer.AddSpacer(10)
        self.westSizer.Add(wx.StaticText(self, label="Label:"), 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.westSizer.AddSpacer(5)
        self.wLabelCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(100,25))
        self.westSizer.Add(self.wLabelCtrl)
        self.mainSizer.Add(self.westSizer)
        self.mainSizer.AddSpacer(10)

        self.buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.buttonSizer.AddSpacer(30)
        self.applyButton  = wx.Button(self, wx.ID_ANY, "Apply", size=(55,25))
        self.applyButton.Bind(wx.EVT_BUTTON, self.OnApply)
        self.buttonSizer.Add(self.applyButton)
        self.buttonSizer.AddSpacer(10)
        self.cancelButton  = wx.Button(self, wx.ID_ANY, "Cancel", size=(55,25))
        self.cancelButton.Bind(wx.EVT_BUTTON, self.OnCancel)
        self.buttonSizer.Add(self.cancelButton)
        self.mainSizer.Add(self.buttonSizer)

        self.mainSizer.Layout()
        self.SetSizerAndFit(self.mainSizer)
        self.SetupScrolling()
        self.Refresh()

    def OnApply(self, event):
        if self.tileType is None:
            return
        newTileType = self.tileType.Copy()
        newTileType.name  = self.nameCtrl.GetValue()
        newTileType.label = self.labelCtrl.GetValue()
        newTileType.sides[N].strength = self.nStrengthCtrl.GetValue()
        newTileType.sides[N].label    = self.nLabelCtrl.GetValue()
        newTileType.sides[E].strength = self.eStrengthCtrl.GetValue()
        newTileType.sides[E].label    = self.eLabelCtrl.GetValue()
        newTileType.sides[S].strength = self.sStrengthCtrl.GetValue()
        newTileType.sides[S].label    = self.sLabelCtrl.GetValue()
        newTileType.sides[W].strength = self.wStrengthCtrl.GetValue()
        newTileType.sides[W].label    = self.wLabelCtrl.GetValue()

        self.origTileColor = self.tileType.color
        self.origTextColor = self.tileType.textColor

        # Verify that the new tile definition is "ok" (i.e. doesn't have same name or same glues as another)
        bNewName = not (newTileType.name == self.tileType.name)
        for ttype in self.frame.canvas.tileTypes:
            if bNewName == True:   # The name has changed - see if it collides with another tile's name
                if newTileType.name == ttype.name:
                    msg = "Error: tile type name '" + newTileType.name + "' already exists."
                    dlg = wx.MessageDialog(self.frame, msg, caption="Error", style=wx.ICON_ERROR|wx.OK)
                    dlg.ShowModal()
                    dlg.Destroy()
                    return
            if (not ttype.name == self.tileType.name) and (ttype.HasMatchingGlues(newTileType) == True):
                msg = "Warning: tile type '" + ttype.name + "' has identical glues.  Click 'OK' to continue with this definition."
                dlg = wx.MessageDialog(self.frame, msg, caption="Warning", style=wx.ICON_WARNING|wx.OK|wx.CANCEL)
                bCancel = (dlg.ShowModal() == wx.ID_CANCEL)
                dlg.Destroy()
                if bCancel == True:
                    return
        
        newTileType.ResetImages()

        # Now copy back to main editor tile set
        for ttype in self.frame.canvas.tileTypes:
            if ttype.name == self.tileType.name:
                ttype.SetEqual(newTileType)
                break
        self.frame.canvas.Refresh()
        
        self.tileType = newTileType
        self.Refresh()


    def OnCancel(self, event):
        if self.tileType is None:
            return
        savedOrigColor = self.origTileColor
        savedOrigTextColor = self.origTextColor
        self.SetTileType(self.tileType)
        self.tileType.color = savedOrigColor
        self.tileType.textColor = savedOrigTextColor
        self.SetTileType(self.tileType)
        self.tileType.ResetImages()
        self.Refresh()

    def SetTileType(self, tileType):
        if tileType is None:
            self.tileType = None
            self.nameCtrl.SetValue("")
            self.labelCtrl.SetValue("")
            self.nStrengthCtrl.SetValue(0)
            self.nLabelCtrl.SetValue("")
            self.eStrengthCtrl.SetValue(0)
            self.eLabelCtrl.SetValue("")
            self.sStrengthCtrl.SetValue(0)
            self.sLabelCtrl.SetValue("")
            self.wStrengthCtrl.SetValue(0)
            self.wLabelCtrl.SetValue("")
            self.tileColorButton.SetBackgroundColour(wx.NullColour)
            self.textColorButton.SetBackgroundColour(wx.NullColour)
            self.origTileColor = None
            self.origTextColor = None
            #self.applyButton.SetLabel("Create")
            
        else:
            self.tileType = tileType.Copy()
            self.nameCtrl.SetValue(self.tileType.GetName())
            self.labelCtrl.SetValue(self.tileType.label)
            self.nStrengthCtrl.SetValue(self.tileType.sides[N].strength)
            self.nLabelCtrl.SetValue(self.tileType.sides[N].label)
            self.eStrengthCtrl.SetValue(self.tileType.sides[E].strength)
            self.eLabelCtrl.SetValue(self.tileType.sides[E].label)
            self.sStrengthCtrl.SetValue(self.tileType.sides[S].strength)
            self.sLabelCtrl.SetValue(self.tileType.sides[S].label)
            self.wStrengthCtrl.SetValue(self.tileType.sides[W].strength)
            self.wLabelCtrl.SetValue(self.tileType.sides[W].label)
            self.origTileColor = self.tileType.color
            self.origTextColor = self.tileType.textColor
            #self.applyButton.SetLabel("Apply")
            
            self.tileType.ParseColor()

            if self.tileType.color is None:
                self.tileColorButton.SetBackgroundColour(wx.NullColour)
            else:
                self.tileColorButton.SetBackgroundColour(self.tileType.color)
                
            if self.tileType.textColor is None:
                self.textColorButton.SetBackgroundColour(wx.NullColour)
            else:
                self.textColorButton.SetBackgroundColour(self.tileType.textColor)
        
        self.Refresh()
        
    def OnTileColor(self, event):
        newColor = self.PickColor(self.tileType.color)
        if newColor is wx.NullColour:
            return
        self.tileType.color = "RGBA" + str(newColor)
        self.tileType.ParseColor()
        self.tileColorButton.SetBackgroundColour(self.tileType.colorVector)
        self.tileType.ResetImages()
        self.Refresh()
        
    def OnTextColor(self, event):
        newColor = self.PickColor(self.tileType.textColor)
        if newColor is wx.NullColour:
            return
        self.tileType.textColor = "RGBA" + str(newColor)
        self.tileType.ParseColor()
        self.textColorButton.SetBackgroundColour(self.tileType.textColorVector)
        self.tileType.ResetImages()
        self.Refresh()
        
    def PickColor(self, currColor):
        retColor = wx.NullColour
        if self.frame.useAdvancedColorPicker == False:
            cd = wx.ColourData()
            cd.SetColour(wx.Colour(currColor))
            dlg = wx.ColourDialog(self, cd)
     
            # Ensure the full colour dialog is displayed, not the abbreviated version.
            dlg.GetColourData().SetChooseFull(True)

            if dlg.ShowModal() == wx.ID_OK:
                data = dlg.GetColourData()
                retColor = data.GetColour().Get()
            dlg.Destroy()
        else:
            cd = wx.ColourData()
            #if type(currColor) is str:
            #    cd.SetColour(wx.TheColourDatabase.Find(currColor))
            #else:
            cd.SetColour(wx.Colour(currColor))
            dlg = CCD.CubeColourDialog(self, cd)
            if dlg.ShowModal() == wx.ID_OK:
                colourData = dlg.GetColourData()
                colorTuple = (colourData.GetColour().Red(), colourData.GetColour().Green(), colourData.GetColour().Blue(), colourData.GetColour().Alpha())
                retColor = colorTuple
            dlg.Destroy()
        return retColor        
            
    def OnPaint(self, event):
        dc = wx.PaintDC(self)
        bg = wx.Brush(self.frame.canvas.backgroundColor)
        #dc.SetBackground(bg)
        #dc.Clear()

        if self.tileType is None:
            return
        
        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        dc.SetUserScale(2.0, 2.0)
        self.tileType.DrawAtLocation(20, 5, dc)


############################################################################################
#
#
class TileTypeCanvas(wx.ScrolledWindow):
    def __init__(self, parent, idIn, frame, size, app, mouseOverCallback=None, leftClickCallback=None):
        wx.ScrolledWindow.__init__(self, parent, idIn, (0, 0), size=size, style=wx.SUNKEN_BORDER)

        self.frame = frame
        self.app = app
        self.drawScale = 1.0
        self.sideLen = CELLSIZE
        self.spacerHoriz = self.sideLen / 2
        self.spacerVert = self.sideLen / 2
        self.bufferTop = 10
        self.bufferLeft = 10
        self.imgOffsetHoriz = 0
        self.imgOffsetVert = 0
        self.backgroundColor = wx.Colour((255, 255, 128, 255))
        #self.backgroundColor = wx.TheColourDatabase.Find("White")
        self.backgroundBrush = wx.Brush(self.backgroundColor)
        self.backgroundPen = wx.Pen(self.backgroundColor)
        self.tileTypes = []
        self.mouseMode = SIMFW_MODE_PAN_ZOOM
        self.dragStart = None
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)
        self.grabCursorImg = wx.Image("images/grab.gif", wx.BITMAP_TYPE_GIF)
        
        self.mouseOverCallback = mouseOverCallback
        self.leftClickCallback = leftClickCallback
        
        self.selectedTile    = None
        self.selectedTileLoc = None
        self.searchedTiles   = None
        
        self.bConfirmDelete = True

        self.contextMenuLoc = None
        self.rubberBand = SimCanvasRubberband.SimCanvasRubberband(drawingSurface=self)
        self.rubberBand.reset(aspectRatio=0.5)
        self.rubberBand.reset()
        
        self.SetVirtualSize(1000,1000)
        self.sideLen = CELLSIZE
        self.UpdateSizeVars()
        self.SetScrollRate(1,1)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        #self.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        #self.Bind(wx.EVT_CONTEXT_MENU, self.OnMouse)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_SIZE, self.SetSize)
        
    def SetSize(self, event):
        self.UpdateSizeVars()

    def UpdateSizeVars(self):
        if len(self.tileTypes) == 0:
            self.numCols = 1
            self.numRows = 1
        else:
            s = self.GetSize()
            vs = self.GetVirtualSize()
            self.numCols = int(s[0] / ((self.sideLen + self.spacerHoriz) * self.drawScale))
            if self.numCols < 1:
                self.numCols = 1
            if self.numCols > len(self.tileTypes):
                self.numCols = len(self.tileTypes)
            self.numRows = (len(self.tileTypes) / self.numCols) + 1

        newWidth = (self.numCols * (self.sideLen + self.spacerHoriz) + self.bufferLeft) * self.drawScale
        newHeight = (self.numRows * (self.sideLen + self.spacerVert) + self.bufferTop) * self.drawScale
        self.SetVirtualSize((newWidth,newHeight))
        self.Refresh()

    def SetTileTypes(self, tileTypes, bNewCopies=True):
        self.tileTypes = []
        for tt in tileTypes.keys():
            if bNewCopies == True:
                self.tileTypes.append(tileTypes[tt].Copy())
            else:
                self.tileTypes.append(tileTypes[tt])
        self.UpdateSizeVars()

    def OnPaint(self, event):

        dc = wx.PaintDC(self)

        bg = wx.Brush(self.backgroundColor)
        dc.SetBackground(bg)
        dc.Clear()

        if len(self.tileTypes) == 0:
            return
        
        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        clipped = dc.GetClippingBox()
        # On OS X, PrepareDC doesn't seem to account for scrolling as it does on Windows
        if "wxOSX" in wx.PlatformInfo:
            xView,yView = self.GetViewStart()
            xDelta,yDelta = self.GetScrollPixelsPerUnit()
            x = clipped[0] + xView * xDelta
            y = clipped[1] + yView * yDelta
        else:
            x = clipped[0]
            y = clipped[1]
        
        #topLeft = self.CoordsFromPixelPos(x, y)
        #bottomRight = self.CoordsFromPixelPos(x+clipped[2], y+clipped[3])
        
        zerozero = self.GetPixelsForCoords(0,0)
        
        #print("in OnPaint, bufferTop: " + str(self.bufferTop) + ", drawScale: " + str(self.drawScale) + ", (0,0): " + str(zerozero))
        
        dc.SetUserScale(self.drawScale, self.drawScale)

        #for i in range(topLeft[0], bottomRight[0]+1):
        #    for j in range(bottomRight[1], topLeft[1]+1):
        
        tileList = self.tileTypes
        if not self.searchedTiles is None:
            tileList = self.searchedTiles
        n = 0
        for row in range(int(self.numRows)):
            for col in range(int(self.numCols)):
                if n < len(tileList):
                    ttype = tileList[n]
                    if ttype == self.selectedTile:
                        self.HighlightSelectedTile(dc, col, row, ttype)
                    self.DrawAtCoords(dc, col, row, ttype)
                        
                n = n + 1
            
        return
    
    def HighlightSelectedTile(self, dc, xCoord, yCoord, ttype, bRemoveHighlight=False):
        savedBrush = dc.GetBrush()
        (x,y) = self.GetPixelsForCoords(xCoord, yCoord)
        if bRemoveHighlight == False:
            dc.SetBrush(wx.Brush(wx.Colour("Purple")))
        else:
            dc.SetBrush(self.backgroundBrush)
        dc.SetPen(self.backgroundPen)
        dc.DrawRectangle((x-self.spacerHoriz/2), (y-self.spacerVert/2), (self.sideLen + self.spacerHoriz), (self.sideLen + self.spacerVert))
        dc.SetBrush(savedBrush)
        self.DrawAtCoords(dc, xCoord, yCoord, ttype)
        
    def GetPixelsForCoords(self, xCoord, yCoord):
        xPixel = xCoord * (self.sideLen + self.spacerHoriz) + self.bufferLeft
        yPixel = yCoord * (self.sideLen + self.spacerVert) + self.bufferTop
        return (xPixel,yPixel)
  
    def CoordsFromPixelPos(self, x, y, bExact=False):
        #if (x % ((self.sideLen + self.spacerHoriz) * self.drawScale)) < (self.sideLen * self.drawScale):
        #    print(str(x))
        if x > self.bufferLeft:
            x = x + self.bufferLeft
        if y > self.bufferTop:
            y = y + self.bufferTop
        col = int(x / ((self.sideLen + self.spacerHoriz) * self.drawScale))
        row = int(y / ((self.sideLen + self.spacerVert) * self.drawScale))
        return (col,row)
        
    def TileFromPixelPos(self, x, y):
        if (self.tileTypes is None) or (len(self.tileTypes) == 0):
            return None
        (col,row) = self.CoordsFromPixelPos(x,y, bExact=True)
        num = (row * self.numCols) + col
        if not self.searchedTiles is None:
            if num < len(self.searchedTiles):
                return self.searchedTiles[num]
        else:
            if num < len(self.tileTypes):
                return self.tileTypes[num]
        return None
    
    def DrawAtCoords(self, dc, xCoord, yCoord, ttype):
        if ttype is None:
            return
        else:
            (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)
            # Adjust coords for correct offset of tile images
            xPixel = xPixel - self.imgOffsetHoriz
            yPixel = yPixel - self.imgOffsetVert
            ttype.DrawAtLocation(xPixel, yPixel, dc)
    
    def GetCenterCoords(self):
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
        coords = self.CoordsFromPixelPos(center[0], center[1])
        return coords

    def ConvertEventCoords(self, event, pixels=None):
        if pixels == None:
            pixels = (event.GetX(), event.GetY())
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (pixels[0] + (xView * xDelta),
                pixels[1] + (yView * yDelta))

    def OnSelectClick(self, xPixel, yPixel):
        #(xPixel, yPixel) = self.ConvertEventCoords(event=None, pixels=(xPixel, yPixel))
        ttype = self.TileFromPixelPos(xPixel, yPixel)
        print("(xPixel, yPixel)" + str((xPixel, yPixel)) + ", ttype: " + str(ttype))
        if not ttype is None:
            if ttype == self.selectedTile:
                dc = wx.ClientDC(self)
                self.HighlightSelectedTile(dc, self.selectedTileLoc[0], self.selectedTileLoc[1], ttype, True)
                self.selectedTile = None
                self.selectedTileLoc = None
            else:
                dc = wx.ClientDC(self)
                if not self.selectedTile is None:
                    self.HighlightSelectedTile(dc, self.selectedTileLoc[0], self.selectedTileLoc[1], self.selectedTile, True)
                self.selectedTile = ttype
                (col,row) = self.CoordsFromPixelPos(xPixel, yPixel)
                self.selectedTileLoc = (col,row)
                self.HighlightSelectedTile(dc, col, row, ttype)
        elif not self.selectedTile is None:
            dc = wx.ClientDC(self)
            self.HighlightSelectedTile(dc, self.selectedTileLoc[0], self.selectedTileLoc[1], self.selectedTile, True)
            self.selectedTile = None
            self.selectedTileLoc = None
        self.Refresh()
        self.frame.tileTypePanel.SetTileType(self.selectedTile)

    def OnMouse(self, event):
        #if type(event) is wx.ContextMenuEvent:
        #    print("got one")

        x,y = self.ConvertEventCoords(event)
        coords = self.CoordsFromPixelPos(x, y)
        ttype = self.TileFromPixelPos(x, y)

        if not self.mouseOverCallback is None:
            self.mouseOverCallback(ttype, coords)
            
        if event.LeftDown():
            if not self.leftClickCallback is None:
                self.leftClickCallback(ttype, coords)

        if event.Leaving():
            #print("leaving")
            self.frame.SetFocus()

    def CanvasSizeUpdate(self, centerCoords=None):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "sizeEditorCanvas", coords = centerCoords)
        
    def DoCanvasSizeUpdate(self, coords=None):
        #If we zoomed, we already changed the scale so can't find the center coords now
        #if coords is None:
        #    coords = self.GetCenterCoords()
        self.UpdateSizeVars()
        if not coords is None:
            self.ScrollToCoords(coords)
        
    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("=") or keycode == ord("+"):
            self.DoZoom(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.DoZoom(False)
        elif keycode == wx.WXK_F2:
            self.Refresh()
        
    def OnZoomIn(self, event):
        self.DoZoom(True)

    def OnZoomOut(self, event):
        self.DoZoom(False)
        
    def DoZoom(self, bZoomIn=True, bScrollToCenter=False, scrollToCoords=None):
        if bScrollToCenter == True:
            coords = self.GetCenterCoords()
        else:
            coords = scrollToCoords
        if bZoomIn == True:
            self.drawScale = self.drawScale * 2.0
        else:
            self.drawScale = self.drawScale / 2.0
        print("Scale set to " + str(self.drawScale))
        self.CanvasSizeUpdate(coords)

    def ScrollToCoords(self, coords):
        (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
        x = x - self.sideLen/2
        y = y - self.sideLen/2
        x = x * self.drawScale
        y = y * self.drawScale
        (xs, ys) = self.GetScrollPixelsPerUnit()
        (xsz,ysz) = self.GetSize()
        self.Scroll((x-xsz/2+(self.sideLen+self.spacerHoriz)*self.drawScale)/xs,
                    (y-ysz/2+(self.sideLen+self.spacerVert)*self.drawScale)/ys)

    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, color):
        if not self.brush_cache.has_key(color):
            self.brush_cache[color] = wx.Brush(color)
        return self.brush_cache[color]

    def SetBackgroundColor(self):
        if self.frame.useAdvancedColorPicker == False:
            dlg = wx.ColourDialog(self)
     
            # Ensure the full colour dialog is displayed, not the abbreviated version.
            dlg.GetColourData().SetChooseFull(True)
     
            if dlg.ShowModal() == wx.ID_OK:
                data = dlg.GetColourData()
                self.backgroundColor = wx.Colour(data.GetColour().Get())
                #self.app.canvas.InitializeDrawing(self.app.canvas.pdc)
                #self.tileTypes.CreateTileTypeImages(self.backgroundColor)
                self.Refresh()
            dlg.Destroy()
        else:
            cd = wx.ColourData()
            if type(self.backgroundColor) is str:
                cd.SetColour(wx.TheColourDatabase.Find(self.backgroundColor))
            else:
                cd.SetColour(self.backgroundColor)
            dlg = CCD.CubeColourDialog(self, cd)
            if dlg.ShowModal() == wx.ID_OK:
                colourData = dlg.GetColourData()
                colorTuple = (colourData.GetColour().Red(), colourData.GetColour().Green(), colourData.GetColour().Blue(), colourData.GetColour().Alpha())
                self.backgroundColor = wx.Colour(colorTuple)
                #box.beadType.color = str(colourData.GetColour().Get())
                #self.app.canvas.InitializeDrawing(self.app.canvas.pdc)
                #self.tileTypes.CreateTileTypeImages(self.backgroundColor)
                self.Refresh()
            dlg.Destroy()
        self.backgroundBrush = wx.Brush(self.backgroundColor)
        self.backgroundPen = wx.Pen(self.backgroundColor)


############################################################################################
#
#
class TileTypePopup(wx.PopupTransientWindow):
    def __init__(self, parent, tileTypes, frame, app, size, assemblyCoords):
        wx.PopupTransientWindow.__init__(self, parent, flags=wx.BORDER_DOUBLE)
        self.parent = parent
        self.frame = frame
        self.tileTypes = tileTypes
        self.assemblyCoords = assemblyCoords
        
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.AddSpacer(5)
        
        self.topSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.topSizer.AddSpacer(3)
        self.tileNameText = wx.StaticText(self, label="Add tile of type: ")
        self.topSizer.Add(self.tileNameText, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.nameCtrl = wx.TextCtrl(self, -1, "", wx.DefaultPosition, wx.Size(180,25), style=wx.TE_PROCESS_ENTER)
        self.nameCtrl.Bind(wx.EVT_TEXT_ENTER, self.OnEnter)
        self.topSizer.Add(self.nameCtrl)
        self.topSizer.AddSpacer(10)
        self.cancelButton = wx.Button(self, wx.ID_ANY, "X", size=(20,25))
        self.cancelButton.Bind(wx.EVT_BUTTON, self.OnCancel)
        self.topSizer.Add(self.cancelButton, 0)
        self.mainSizer.Add(self.topSizer)
        self.mainSizer.AddSpacer(5)
        
        self.canvas = TileTypeCanvas(self, -1, frame, size, app, mouseOverCallback=self.SetCurrentTileType, leftClickCallback=self.OnTileSelect)
        self.canvas.SetTileTypes(self.tileTypes, bNewCopies=False)
        
        self.mainSizer.Add(self.canvas)

        self.mainSizer.Layout()
        self.SetSizerAndFit(self.mainSizer)
        
    def SetCurrentTileType(self, ttype, coords):
        if not ttype is None:
            self.nameCtrl.SetLabel(ttype.GetName())
        
    def OnTileSelect(self, ttype, coords):
        if (not ttype is None) and (not self.assemblyCoords is None):
            self.parent.assembly.AddTile(ttype, self.assemblyCoords)
            self.parent.Refresh()
            self.frame.SetFocus()
            self.Dismiss()
            
    def OnEnter(self, event):
        #print("enter was pressed")
        return

    def OnCancel(self, event):
        self.Dismiss()


############################################################################################
#
#
class DeleteTilePopup(wx.PopupTransientWindow):
    def __init__(self, parent, tileType, frame, app, size, assemblyCoords):
        wx.PopupTransientWindow.__init__(self, parent, flags=wx.BORDER_DOUBLE)
        self.parent = parent
        self.frame = frame
        self.tileType = tileType
        self.assemblyCoords = assemblyCoords
        
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.AddSpacer(5)
        
        self.topSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.topSizer.AddSpacer(3)
        self.tileNameText = wx.StaticText(self, label="Delete tile at " + str(self.assemblyCoords) + "?")
        self.topSizer.Add(self.tileNameText, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        self.topSizer.AddSpacer(5)
        self.okButton = wx.Button(self, wx.ID_ANY, "OK", size=(25,25))
        self.okButton.Bind(wx.EVT_BUTTON, self.OnOk)
        self.topSizer.Add(self.okButton, 0, wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
        self.topSizer.AddSpacer(5)
        self.cancelButton = wx.Button(self, wx.ID_ANY, "Cancel", size=(60,25))
        self.cancelButton.Bind(wx.EVT_BUTTON, self.OnCancel)
        self.topSizer.Add(self.cancelButton, 0, wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)

        self.mainSizer.Add(self.topSizer)
        self.mainSizer.AddSpacer(5)

        self.mainSizer.Layout()
        self.SetSizerAndFit(self.mainSizer)
        
    def OnOk(self, event):
        self.parent.assembly.RemoveTile(self.assemblyCoords)
        self.parent.Refresh()
        self.frame.SetFocus()
        self.Dismiss()
        
    def OnCancel(self, event):
        self.frame.SetFocus()
        self.Dismiss()

############################################################################################
#
#
class TileTypeEditorCanvas(TileTypeCanvas):
    def __init__(self, parent, idIn, frame, size, app, tileTypeClassNameIn):
        TileTypeCanvas.__init__(self, parent, idIn, frame, size, app)
        
        self.tileTypeClassName = tileTypeClassNameIn
        self.SubscribeEventHandlers()
        self.CanvasSizeUpdate()
        
    def SetTileTypes(self, tileTypes, bNewCopies=True):
        TileTypeCanvas.SetTileTypes(self, tileTypes, bNewCopies)
        self.frame.OutputMessage("Loaded " + str(len(self.tileTypes)) + " tile types.")

    def OnMouse(self, event):
        x,y = self.ConvertEventCoords(event)
        coords = self.CoordsFromPixelPos(x, y)
        ttype = self.TileFromPixelPos(x, y)
        if  ttype is None:
            self.SetMouseOverText(None)
            if self.selectedTile is None:
                self.frame.tileTypePanel.SetTileType(None)
        else:
            self.SetMouseOverText(ttype)
            if self.selectedTile is None:
                self.frame.tileTypePanel.SetTileType(ttype)
        
        if event.RightDown():
            self.OnContext(event, x, y)
            return

        if self.mouseMode == SIMFW_MODE_SELECT:
            if event.RightUp():
                self.rubberBand.reset()
            else:
                self.rubberBand.HandleMouseEvents(event)
        else:
            if event.LeftDown():
                if event.ShiftDown():
                    self.OnSelectClick(x, y)
                elif self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                    self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                    self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))
    
            elif event.LeftDClick():
                self.OnSelectClick(x, y)
            elif event.RightDown():
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    
                    self.OnContext(event, x, y)
                    return
                    
                    bZoomIn = True
                    if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                        bZoomIn = False
                    self.DoZoom(bZoomIn=bZoomIn, bScrollToCenter=False, scrollToCoords=coords)
                    #self.ScrollToCoords((x2,y2))
                    sz = self.GetSize()
                    self.WarpPointer(sz[0]/2, sz[1]/2)

            elif event.LeftUp():
                self.dragStart = None
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabCursorImg))
                else:
                    self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            
            elif event.Dragging():
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    if self.dragStart is not None:
                        #self.SetScrollRate(1,1)
                        new = self.ScreenToClient(wx.GetMousePosition())
                        dx = int((self.dragStart.x-new.x)/1)
                        dy = int((self.dragStart.y-new.y)/1)
                        spx = self.GetScrollPos(wx.HORIZONTAL)
                        spy = self.GetScrollPos(wx.VERTICAL)
                        self.Scroll(spx+dx, spy+dy)
                        self.dragStart = new


    def SendStatusBarUpdate(self, field, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateEditorStatusbar", msg = (field, msg))

    def UpdateStatusbar(self, msg):
        self.frame.SetStatusText(msg[1], msg[0])

    def SetMouseOverText(self, ttype):
        if ttype == None:
            self.SendStatusBarUpdate(0, "")
            self.SendStatusBarUpdate(2, "")
            return

        tileTypeName = "Tile type: " + ttype.ToString(bFull=True)
        self.SendStatusBarUpdate(2, tileTypeName)
        
    def SubscribeEventHandlers(self):
        pub.getDefaultPublisher().subscribe(self.UpdateStatusbar, "updateEditorStatusbar")
        #pub.getDefaultPublisher().subscribe(self.DoDrawAssemblyUpdate, "drawAssembly")
        pub.getDefaultPublisher().subscribe(self.DoCanvasSizeUpdate, "sizeEditorCanvas")
        
    def UnsubscribeEventHandlers(self):
        pub.getDefaultPublisher().unsubscribe(self.UpdateStatusbar, "updateEditorStatusbar")
        #pub.getDefaultPublisher().unsubscribe(self.DoDrawAssemblyUpdate, "drawAssembly")
        pub.getDefaultPublisher().unsubscribe(self.DoCanvasSizeUpdate, "sizeEditorCanvas")
        
    def OnContext(self, event, xPixel, yPixel):
        """
        Create and show a Context Menu
        """
        self.contextMenuLoc = (xPixel, yPixel)

        # Only do this part the first time so the events are only bound once 
        if not hasattr(self, "popupEdit"):
            self.popupEdit      = wx.NewId()
            self.popupDuplicate = wx.NewId()
            self.popupDelete    = wx.NewId()
            self.popupNew       = wx.NewId()
            self.Bind(wx.EVT_MENU, self.OnPopupEdit, id=self.popupEdit)
            self.Bind(wx.EVT_MENU, self.OnPopupDuplicate, id=self.popupDuplicate)
            self.Bind(wx.EVT_MENU, self.OnPopupDelete, id=self.popupDelete)
            self.Bind(wx.EVT_MENU, self.OnPopupNew,   id=self.popupNew)
 
        # Build the menu
        menu = wx.Menu()
        item = menu.Append(self.popupEdit, "Edit")
        item = menu.Append(self.popupDuplicate, "Duplicate")
        item = menu.Append(self.popupDelete, "Delete")
        item = menu.Append(self.popupNew, "New")
 
        # Show the popup menu
        self.PopupMenu(menu)
        menu.Destroy()
        self.contextMenuLoc = None
        
    def OnPopupEdit(self, event):
        self.OnSelectClick(self.contextMenuLoc[0], self.contextMenuLoc[1])
        
    def OnPopupDuplicate(self, event):
        #x,y = self.ConvertEventCoords(event)
        #coords = self.CoordsFromPixelPos(self.contextMenuLoc[0], self.contextMenuLoc[1])
        ttype = self.TileFromPixelPos(self.contextMenuLoc[0], self.contextMenuLoc[1])
        if ttype == None:
            return
        tTypeCopy = ttype.Copy()
        ttypeName = tTypeCopy.GetName() + "-Copy"
        num = 1
        bKeepGoing = True
        while bKeepGoing == True:
            bKeepGoing = False
            for tt in self.tileTypes:
                if ttypeName == tt.GetName():
                    num = num + 1
                    ttypeName = tTypeCopy.GetName() + "-Copy" + str(num)
                    bKeepGoing = True
        tTypeCopy.name = ttypeName
                
        self.tileTypes.append(tTypeCopy)
        self.Refresh()
        
    def OnPopupDelete(self, event):
        ttype = self.TileFromPixelPos(self.contextMenuLoc[0], self.contextMenuLoc[1])
        if ttype is None:
            return
        bDeleteIt = False
        if self.bConfirmDelete == True:
            dlg = SimUtils.DontShowAgainDialog(self, wx.NewId(), "Confirm delete", "Are you sure you want to delete tile type '" + ttype.GetName() + "'?")
            retVal = dlg.ShowModal()
            if retVal == SimUtils.SIMU_ID_OK_CHECK:
                self.bConfirmDelete = False
                bDeleteIt = True
            elif retVal == SimUtils.SIMU_ID_OK_NOCHECK:
                bDeleteIt = True
        else:
            bDeleteIt = True
        if bDeleteIt == True:
            index = 0
            for tt in self.tileTypes:
                if tt == ttype:
                    break
                index = index + 1
            self.OnSelectClick(self.contextMenuLoc[0], self.contextMenuLoc[1])
            del(self.tileTypes[index])
            self.Refresh()

    def OnPopupNew(self, event):
        ttypeName = "NewTile"
        newTType = self.tileTypeClassName(ttypeName, label="")
        num = 1
        bKeepGoing = True
        while bKeepGoing == True:
            bKeepGoing = False
            for tt in self.tileTypes:
                if ttypeName == tt.GetName():
                    num = num + 1
                    ttypeName = "NewTile" + str(num)
                    bKeepGoing = True
        newTType.name = ttypeName
        self.tileTypes.append(newTType)
        self.UpdateSizeVars()
        self.Refresh()


############################################################################################
#
#
class TileTypeEditorFrame(wx.Frame):

    def __init__(self, app, tileTypeClassName):
        self.app = app
        title = "Tile Type Editor"
        wx.Frame.__init__(self, None, -1, title, pos=(50,50), size=(1200,800),
                         style=wx.DEFAULT_FRAME_STYLE, name=title)
        
        self.tileTypePanel = None
        self.useAdvancedColorPicker = False
        self.tileSetFile = None
        self.tileTypeClassName = tileTypeClassName

        self._mgr = wx.aui.AuiManager(self)
        self.CreateMenu()
        self.Show(True)
        
        self.CreateCanvas(self.app)
        self.MakeStatusBar()
        self.CreateToolbar()
        self.AddTileTypeWindow()
        self.AddMessageWindow()

        self._mgr.AddPane(self.canvas, wx.CENTER, 'Tile Type Editor')
        
        self.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        
        self.SetSize((1200, 800))
        
        # tell the manager to 'commit' all the changes just made
        self._mgr.Update()
        
    def OutputMessage(self, msg):
        self.messageWin.VisibleWrite(msg + "\n")
        
    def AddTileTypeWindow(self):
        # Destroy the existing copy of the window
        if not self.tileTypePanel is None:
            pane = self._mgr.GetPane(self.tileTypePanel)
            pane.Hide()
            self._mgr.DetachPane(self.tileTypePanel)
            self._mgr.Update()
            #for bb in self.beadBoxes:
            #    self.beadBoxes[bb].Destroy()
            self.tileTypePanel.Destroy()
            self.beadTypePanel = None
        self.tileTypePanel = TileTypePanel(self)
        tileTypePanelSizer = wx.GridSizer(cols=1, hgap=5, vgap=5)
        
#===============================================================================
#         self.beadBoxes = dict()
#         bb = OritatamiBeadBox(None, self.beadTypePanel, beadTypePanelSizer, self, self.beadBoxes)
#         self.beadBoxes[""] = bb
# 
#         for beadTypeName in self.canvas.assembly.ONodeTypes:
#             beadType = self.canvas.assembly.ONodeTypes[beadTypeName]
#             bb = OritatamiBeadBox(beadType, self.beadTypePanel, beadTypePanelSizer, self, self.beadBoxes)
#             self.beadBoxes[beadTypeName] = bb
#         beadTypePanelSizer.Layout()
#===============================================================================

        
        self.tileTypePanel.SetSizerAndFit(tileTypePanelSizer)
        
        #self.tileTypePanel.Bind(wx.EVT_MOUSE_EVENTS, self.OnTilePanelClick)
        #self.beadTypePanel.Bind(wx.EVT_SIZE, self.OnTilePanelResize)
        #self._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseTileEditPane)
        self._mgr.AddPane(self.tileTypePanel, wx.aui.AuiPaneInfo().Left().BestSize((TILEPANEL_DEFWIDTH,TILEPANEL_DEFHEIGHT)).FloatingSize((TILEPANEL_DEFWIDTH,TILEPANEL_DEFHEIGHT)).Caption('Tile Type Window'))
        self._mgr.Update()
        self.tileTypePanel.SetupScrolling()
        self.ttPanelSize = self.tileTypePanel.GetSize()
        
        #self.menuView.Check(self.itemShowBTEdit.GetId(), True)
        
    def AddMessageWindow(self):
        self.ctrlPanel = wx.Panel(self, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,200), style = wx.TAB_TRAVERSAL)
        self.messageWin = SimUtils.SimRTC(self.ctrlPanel, self, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,200))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        sz = wx.BoxSizer(wx.HORIZONTAL)
        sz.Add(self.messageWin, 20, wx.EXPAND, 20)
        self.ctrlPanel.SetSizerAndFit(sz)
        #self._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseMessagePane)
        self._mgr.AddPane(self.ctrlPanel, wx.aui.AuiPaneInfo().Bottom().BestSize((1000,100)).FloatingSize((1000,100)).Caption('Message Panel'))
        self._mgr.Update()
        
    def CreateToolbar(self):
        self.toolbar = wx.ToolBar(self, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        self.toolbar.SetToolSeparation(20)
        
        self.searchCtrl = EditorSearchCtrl(self.toolbar, self, size=(200,-1), doSearch=self.DoSearch)
        self.toolbar.AddControl(self.searchCtrl)
        self.searchCtrl.Bind(wx.EVT_SEARCHCTRL_SEARCH_BTN, self.DoSearch)
        self.toolbar.AddSeparator()
        
        self.radioExact = wx.RadioButton(self.toolbar, -1, "Exact", style=wx.RB_GROUP)
        self.radioExact.SetValue(True)
        self.toolbar.AddControl(self.radioExact)
        self.radioPartial = wx.RadioButton(self.toolbar, -1, "Partial")
        self.toolbar.AddControl(self.radioPartial)

        self.searchName = wx.CheckBox(self.toolbar, id=-1, label="Name")
        self.searchName.SetValue(True)
        self.toolbar.AddControl(self.searchName)
        self.searchLabel = wx.CheckBox(self.toolbar, id=-1, label="Label")
        self.toolbar.AddControl(self.searchLabel)
        self.searchNorth = wx.CheckBox(self.toolbar, id=-1, label="North")
        self.toolbar.AddControl(self.searchNorth)
        self.searchEast = wx.CheckBox(self.toolbar, id=-1, label="East")
        self.toolbar.AddControl(self.searchEast)
        self.searchSouth = wx.CheckBox(self.toolbar, id=-1, label="South")
        self.toolbar.AddControl(self.searchSouth)
        self.searchWest = wx.CheckBox(self.toolbar, id=-1, label="West")
        self.toolbar.AddControl(self.searchWest)
        self.toolbar.AddSeparator()

        #self.toolbar.SetToolBitmapSize(wx.Size(24,24))
        #self.toolbar.AddTool(SIMID_ED_SEARCH, "Fast Rewind", wx.Image("images/fastleft.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Rewind")
        self.toolbar.Realize()
        #self.toolbar.Bind(wx.EVT_TOOL, self.OnTool)
        self._mgr.AddPane(self.toolbar, wx.aui.AuiPaneInfo().
                          Name("Editor toolbar").Caption("Editor Toolbar").
                          ToolbarPane().Top())

    def CreateMenu(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        itemNew = menu.Append(-1, "&New tile set\tCtrl-N", "New tile set")
        self.Bind(wx.EVT_MENU, self.OnNew, itemNew)
        menu.AppendSeparator()
        item1 = menu.Append(-1, "&Load tile set...\tCtrl-L", "Load tile set...")
        self.Bind(wx.EVT_MENU, self.OnLoad, item1)
        itemR = menu.Append(-1, "&Reload tile set from file\tCtrl-R", "Reload tile set from file")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemLoadFromSim = menu.Append(-1, "Load tile set from simulator", "Load tile set from simulator")
        self.Bind(wx.EVT_MENU, self.OnLoadFromSim, itemLoadFromSim)
        menu.AppendSeparator()
        itemSaveAs = menu.Append(-1, "Save tile set as...", "Save tile set as...")
        self.Bind(wx.EVT_MENU, self.OnSaveAs, itemSaveAs)
        itemSave = menu.Append(-1, "&Save tile set\tCtrl-S", "Save tile set")
        self.Bind(wx.EVT_MENU, self.OnSave, itemSave)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit editor")
        self.Bind(wx.EVT_MENU, self.OnExit, item3)
        menuBar.Append(menu, "&File")
        
        self.menuView = wx.Menu()
        itemZIn = self.menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = self.menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        self.menuView.AppendSeparator()
        itemBGC = self.menuView.Append(-1, "Set background color", "Set background color")
        self.Bind(wx.EVT_MENU, self.OnSetBackgroundColor, itemBGC)
        self.menuView.AppendSeparator()
        itemRefresh = self.menuView.Append(-1, "Refresh screen\tF2", "Refresh screen")
        self.Bind(wx.EVT_MENU, self.OnRefresh, itemRefresh)

        menuBar.Append(self.menuView, "&View")
        
        self.menuSettings = wx.Menu()
        self.itemAdvColor = self.menuSettings.Append(-1, "Use advanced color picker", "Use advanced color picker dialog", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnAdvancedColorChecked, self.itemAdvColor)
        self.menuSettings.Check(self.itemAdvColor.GetId(), False)
        menuBar.Append(self.menuSettings, "&Settings")
        
        self.SetMenuBar(menuBar)
        
    def OnZoomIn(self, event):
        self.canvas.OnZoomIn(event)

    def OnZoomOut(self, event):
        self.canvas.OnZoomOut(event)
        
    def OnSetBackgroundColor(self, event):
        self.canvas.SetBackgroundColor()
        
    def OnRefresh(self, event):
        self.canvas.Refresh()

    def OnAdvancedColorChecked(self, event):
        if self.itemAdvColor.IsChecked():
            self.useAdvancedColorPicker = True
        else:
            self.useAdvancedColorPicker = False

    def MakeStatusBar(self):
        self.CreateStatusBar(3)
        self.SetStatusWidths([150,200,-1])

    def CreateCanvas(self, app):
        self.canvas = TileTypeEditorCanvas(self, wx.ID_ANY, self, wx.Size(500,500), app, self.tileTypeClassName)

    def DoSearch(self, event):
        hitList = []
        searchText = self.searchCtrl.GetValue()
        if len(searchText) > 0:
            for ttype in self.canvas.tileTypes:
                bHit = False
                if self.radioExact.GetValue() == True:
                    if self.searchName.IsChecked() and (not ttype.name is None) and (searchText == ttype.name):
                        bHit = True
                    if self.searchLabel.IsChecked() and (not ttype.label is None) and (searchText == ttype.label):
                        bHit = True
                    if self.searchNorth.IsChecked() and (not ttype.sides[N].label is None) and (searchText == ttype.sides[N].label):
                        bHit = True
                    if self.searchEast.IsChecked() and (not ttype.sides[E].label is None) and (searchText == ttype.sides[E].label):
                        bHit = True
                    if self.searchSouth.IsChecked() and (not ttype.sides[S].label is None) and (searchText == ttype.sides[S].label):
                        bHit = True
                    if self.searchWest.IsChecked() and (not ttype.sides[W].label is None) and (searchText == ttype.sides[W].label):
                        bHit = True
                else:
                    if self.searchName.IsChecked() and (not ttype.name is None) and (searchText in ttype.name):
                        bHit = True
                    if self.searchLabel.IsChecked() and (not ttype.label is None) and (searchText in ttype.label):
                        bHit = True
                    if self.searchNorth.IsChecked() and (not ttype.sides[N].label is None) and (searchText in ttype.sides[N].label):
                        bHit = True
                    if self.searchEast.IsChecked() and (not ttype.sides[E].label is None) and (searchText in ttype.sides[E].label):
                        bHit = True
                    if self.searchSouth.IsChecked() and (not ttype.sides[S].label is None) and (searchText in ttype.sides[S].label):
                        bHit = True
                    if self.searchWest.IsChecked() and (not ttype.sides[W].label is None) and (searchText in ttype.sides[W].label):
                        bHit = True
                if bHit == True:
                    hitList.append(ttype)
            if len(hitList) > 0:
                self.OutputMessage("Search found: " + str(len(hitList)) + " matches.")
            else:
                self.OutputMessage("Search for '" + searchText + "' returned no matches.")
            self.canvas.searchedTiles = hitList
        else:
            self.canvas.searchedTiles = None
        self.canvas.ScrollToCoords((0,0))
        self.SetFocus()
        self.canvas.Refresh()

    def OnTool(self, event):
        print("Sorry, toolbar functions haven't been implemented yet...")

    def OnNew(self, event):
        bDoIt = True
        if len(self.canvas.tileTypes) > 0:
            bDoIt = False
            dlg = wx.MessageDialog(self.canvas, "Are you sure you want to create a new tile set and lose any unsaved changes?", "Confirm new tile set", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
            if dlg.ShowModal() == wx.ID_OK:
                bDoIt = True
            dlg.Destroy()
        if bDoIt == True:
            self.canvas.SetTileTypes(dict())
            self.canvas.tileTypes = None

    def OnLoad(self, event):
        bDoIt = True
        if len(self.canvas.tileTypes) > 0:
            bDoIt = False
            dlg = wx.MessageDialog(self.canvas, "Are you sure you want to load a tile set file and lose any unsaved changes?", "Confirm loading tile set", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
            if dlg.ShowModal() == wx.ID_OK:
                bDoIt = True
            dlg.Destroy()
        if bDoIt == True:
            openFileDialog = wx.FileDialog(self.canvas, "Open", "", "", 
                                           "Tile set files (*.tds)|*.tds",
                                           wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
            openFileDialog.ShowModal()
            fileName = openFileDialog.GetPath()
            openFileDialog.Destroy()
            self.DoLoadTileSet(fileName)
        
    def OnReload(self, event):
        if not self.tileSetFile is None:
            bDoIt = True
            if len(self.canvas.tileTypes) > 0:
                bDoIt = False
                dlg = wx.MessageDialog(self.canvas, "Are you sure you want to load a tile set file and lose any unsaved changes?", "Confirm loading tile set", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
                if dlg.ShowModal() == wx.ID_OK:
                    bDoIt = True
                dlg.Destroy()
            if bDoIt == True:
                self.DoLoadTileSet(self.tileSetFile)
       
    def DoLoadTileSet(self, fileName):
        if (len(fileName) > 0):
            assembly = Assembly()
            loadRet = assembly.LoadTileSet(fileName)
            if loadRet == Assembly.TILESET_LOAD_FAIL_NEEDS_3D:
                wx.MessageDialog(self, "The new tile set requires 3D mode which is not currently supported in the tile type editor.", style=wx.OK|wx.CENTRE|wx.ICON_ERROR).ShowModal()
                return
            elif loadRet == Assembly.TILESET_LOAD_SUCCESS:
                self.tileSetFile = fileName
                self.canvas.SetTileTypes(assembly.tileTypes)

    def OnLoadFromSim(self, event):
        dlg = wx.MessageDialog(self.canvas, "Are you sure you want to load the tile set from the\nsimulator window and lose any unsaved changes?", "Confirm loading tile set", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
        if dlg.ShowModal() == wx.ID_OK:
            self.canvas.SetTileTypes(self.app.canvas.assembly.tileTypes)
        dlg.Destroy()
 
    def OnSaveAs(self, event):
        if (self.canvas.tileTypes is None) or len(self.canvas.tileTypes) < 1:
            wx.MessageDialog(self, "There are no tile types to save.", "No tile types to save", style=wx.OK|wx.CENTRE|wx.ICON_ERROR).ShowModal()
            return
        saveFileDialog = wx.FileDialog(self, "Save tile set as...", "", "", 
                                       "Tile set files (*.tds)|*.tds",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        saveFileDialog.ShowModal()
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        if len(fileName) > 0:
            assembly = Assembly()
            assembly.tileTypes = self.canvas.tileTypes
            if assembly.SaveTileset(fileName) == True:
                self.tileSetFile = fileName
                self.OutputMessage("Successfully saved tile set to file " + str(self.tileSetFile))
            
    def OnSave(self, event):
        if (self.canvas.tileTypes is None) or len(self.canvas.tileTypes) < 1:
            wx.MessageDialog(self, "There are no tile types to save.", "No tile types to save", style=wx.OK|wx.CENTRE|wx.ICON_ERROR).ShowModal()
            return
        if (self.tileSetFile is None) or len(self.tileSetFile) < 1:
            self.OnSaveAs(event)
            return
        assembly = Assembly()
        assembly.tileTypes = self.canvas.tileTypes
        if assembly.SaveTileset(self.tileSetFile) == True:
            self.OutputMessage("Successfully saved tile set to file " + str(self.tileSetFile))

    def OnExit(self, event):
        self.OnCloseFrame(event)

    def OnCloseFrame(self, evt):
        self.app.EditorFrameClosed()
        # deinitialize the frame manager
        self._mgr.UnInit()
        # delete the frame
        self.Destroy()
